local addonName, addonTable = ...

--[[
Setting a value in the DB that matches the existing default will remove the
key from the saved variables.

4 "values" in the DB are actually functions that cannot be used as saved
variables.

- Initialize()
    Intended to be called after ADDON_LOADED, before anything accesses the DB.

- SetDefaults(table)
    Sets the default values to `table`.

	Triggers callbacks for any affected keys.

- RegisterCallback(ident, key, func)
    Calls func(key, oldValue, newValue) every time the key in the DB changes.
    This is not called if the key is assigned its existing value.

    The `ident` value is to allow multiple callbacks for the key.

- UnregisterCallback(ident, key)
    Unregisters the callback for `key` with the given `ident`.

Any callbacks present when the database loads are called immediately, even if
the new value is `nil`. The oldValue is the current default, or `nil`.]]

local defaults = {}
local callbacks = {}

local init

local function dispatchCallbacks(key, oldValue, value)
	if callbacks[key] then
		for ident, func in pairs(callbacks[key]) do
			if func then
				callbacks[key][ident] = false
				xpcall(function() func(key, oldValue, value) end, errorhandler)
				if callbacks[key][ident] ~= nil then
					callbacks[key][ident] = func
				end
			end
		end
	end
end

local funcs
funcs = {
	Initialize = function(self)
		if not init then
			error("Initialize() can only be called once", 2)
		end
		if select(2, IsAddOnLoaded(addonName)) == nil then
			error("Initialize() cannot be called until ADDON_LOADED", 2)
		end
		init()
		init = nil
	end,
	SetDefaults = function(self, table)
		local oldDefaults = defaults
		defaults = table

		for key, t in pairs(callbacks) do
			if init or StormEarthAndFireDB[key] == nil then
				local v, ov = defaults[key], oldDefaults[key]
				if v ~= ov then
					dispatchCallbacks(key, ov, v)
				end
			end
		end
	end,
	RegisterCallback = function(self, ident, key, func)
		if ident == nil then
			error("Identifier must be non-nil", 2)
		elseif type(key) ~= "string" then
			error("Key must be a string", 2)
		elseif funcs[key] then
			error("Key may not be one of the reserved keys", 2)
		elseif type(func) ~= "function" then
			error("Callback must be a function", 2)
		end
		callbacks[key] = callbacks[key] or {}
		callbacks[key][ident] = func
	end,
	UnregisterCallback = function(self, ident, key)
		if callbacks[key] then
			callbacks[key][ident] = nil
		end
	end
}

addonTable.DB = setmetatable({}, {
	__index = function(t, key)
		local v = funcs[key]
		if v == nil then v = defaults[key] end
		return v
	end,
	__newindex = function(t, key, value)
		if funcs[key] then
			error("The key " .. key .. " is reserved")
		end
		error("Saved variables have not yet been loaded")
	end
})

function init()
	local function errorhandler(...)
		return geterrorhandler()(...)
	end


	StormEarthAndFireDB = StormEarthAndFireDB or {}
	setmetatable(addonTable.DB, {
		__index = function(t, key)
			local v = funcs[key]
			if v == nil then v = StormEarthAndFireDB[key] end
			if v == nil then v = defaults[key] end
			return v
		end,
		__newindex = function(t, key, value)
			if funcs[key] then
				error("The key " .. key .. " is reserved")
			end
			local oldValue = StormEarthAndFireDB[key]
			if oldValue == nil then oldValue = defaults[key] end
			if value == defaults[key] then
				StormEarthAndFireDB[key] = nil
			else
				StormEarthAndFireDB[key] = value
			end
			if oldValue ~= value then
				dispatchCallbacks(key, oldValue, value)
			end
		end
	})
	for key in pairs(callbacks) do
		dispatchCallbacks(key, nil, addonTable.DB[key])
	end
end
